# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page
from twilio.rest.preview.trusted_comms.business.brand.branded_channel import BrandedChannelList


class BrandList(ListResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, business_sid):
        """
        Initialize the BrandList

        :param Version version: Version that contains the resource
        :param business_sid: Business Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandList
        """
        super(BrandList, self).__init__(version)

        # Path Solution
        self._solution = {'business_sid': business_sid, }

    def get(self, sid):
        """
        Constructs a BrandContext

        :param sid: Brand Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        """
        return BrandContext(self._version, business_sid=self._solution['business_sid'], sid=sid, )

    def __call__(self, sid):
        """
        Constructs a BrandContext

        :param sid: Brand Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        """
        return BrandContext(self._version, business_sid=self._solution['business_sid'], sid=sid, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.TrustedComms.BrandList>'


class BrandPage(Page):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, response, solution):
        """
        Initialize the BrandPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param business_sid: Business Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandPage
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandPage
        """
        super(BrandPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of BrandInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandInstance
        """
        return BrandInstance(self._version, payload, business_sid=self._solution['business_sid'], )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.TrustedComms.BrandPage>'


class BrandContext(InstanceContext):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, business_sid, sid):
        """
        Initialize the BrandContext

        :param Version version: Version that contains the resource
        :param business_sid: Business Sid.
        :param sid: Brand Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        """
        super(BrandContext, self).__init__(version)

        # Path Solution
        self._solution = {'business_sid': business_sid, 'sid': sid, }
        self._uri = '/Businesses/{business_sid}/Brands/{sid}'.format(**self._solution)

        # Dependents
        self._branded_channels = None

    def fetch(self):
        """
        Fetch the BrandInstance

        :returns: The fetched BrandInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return BrandInstance(
            self._version,
            payload,
            business_sid=self._solution['business_sid'],
            sid=self._solution['sid'],
        )

    @property
    def branded_channels(self):
        """
        Access the branded_channels

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelList
        """
        if self._branded_channels is None:
            self._branded_channels = BrandedChannelList(
                self._version,
                business_sid=self._solution['business_sid'],
                brand_sid=self._solution['sid'],
            )
        return self._branded_channels

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.TrustedComms.BrandContext {}>'.format(context)


class BrandInstance(InstanceResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, payload, business_sid, sid=None):
        """
        Initialize the BrandInstance

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandInstance
        """
        super(BrandInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload.get('account_sid'),
            'business_sid': payload.get('business_sid'),
            'sid': payload.get('sid'),
            'links': payload.get('links'),
            'url': payload.get('url'),
        }

        # Context
        self._context = None
        self._solution = {'business_sid': business_sid, 'sid': sid or self._properties['sid'], }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: BrandContext for this BrandInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandContext
        """
        if self._context is None:
            self._context = BrandContext(
                self._version,
                business_sid=self._solution['business_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: Account Sid.
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def business_sid(self):
        """
        :returns: Business Sid.
        :rtype: unicode
        """
        return self._properties['business_sid']

    @property
    def sid(self):
        """
        :returns: Brand Sid.
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def links(self):
        """
        :returns: Nested resource URLs.
        :rtype: unicode
        """
        return self._properties['links']

    @property
    def url(self):
        """
        :returns: The URL of this resource.
        :rtype: unicode
        """
        return self._properties['url']

    def fetch(self):
        """
        Fetch the BrandInstance

        :returns: The fetched BrandInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandInstance
        """
        return self._proxy.fetch()

    @property
    def branded_channels(self):
        """
        Access the branded_channels

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelList
        """
        return self._proxy.branded_channels

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.TrustedComms.BrandInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import deserialize
from twilio.base import serialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class VerificationList(ListResource):

    def __init__(self, version, service_sid):
        """
        Initialize the VerificationList

        :param Version version: Version that contains the resource
        :param service_sid: The SID of the Service that the resource is associated with

        :returns: twilio.rest.verify.v2.service.verification.VerificationList
        :rtype: twilio.rest.verify.v2.service.verification.VerificationList
        """
        super(VerificationList, self).__init__(version)

        # Path Solution
        self._solution = {'service_sid': service_sid, }
        self._uri = '/Services/{service_sid}/Verifications'.format(**self._solution)

    def create(self, to, channel, custom_friendly_name=values.unset,
               custom_message=values.unset, send_digits=values.unset,
               locale=values.unset, custom_code=values.unset, amount=values.unset,
               payee=values.unset, rate_limits=values.unset,
               channel_configuration=values.unset, app_hash=values.unset):
        """
        Create the VerificationInstance

        :param unicode to: The phone number or email to verify
        :param unicode channel: The verification method to use
        :param unicode custom_friendly_name: A custom user defined friendly name
        :param unicode custom_message: The text of a custom message to use for the verification
        :param unicode send_digits: The digits to send after a phone call is answered
        :param unicode locale: The locale to use for the verification SMS or call
        :param unicode custom_code: A pre-generated code
        :param unicode amount: The amount of the associated PSD2 compliant transaction.
        :param unicode payee: The payee of the associated PSD2 compliant transaction
        :param dict rate_limits: The custom key-value pairs of Programmable Rate Limits.
        :param dict channel_configuration: Channel specific configuration in json format.
        :param unicode app_hash: Your App Hash to be appended at the end of an SMS.

        :returns: The created VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        data = values.of({
            'To': to,
            'Channel': channel,
            'CustomFriendlyName': custom_friendly_name,
            'CustomMessage': custom_message,
            'SendDigits': send_digits,
            'Locale': locale,
            'CustomCode': custom_code,
            'Amount': amount,
            'Payee': payee,
            'RateLimits': serialize.object(rate_limits),
            'ChannelConfiguration': serialize.object(channel_configuration),
            'AppHash': app_hash,
        })

        payload = self._version.create(method='POST', uri=self._uri, data=data, )

        return VerificationInstance(self._version, payload, service_sid=self._solution['service_sid'], )

    def get(self, sid):
        """
        Constructs a VerificationContext

        :param sid: The unique string that identifies the resource

        :returns: twilio.rest.verify.v2.service.verification.VerificationContext
        :rtype: twilio.rest.verify.v2.service.verification.VerificationContext
        """
        return VerificationContext(self._version, service_sid=self._solution['service_sid'], sid=sid, )

    def __call__(self, sid):
        """
        Constructs a VerificationContext

        :param sid: The unique string that identifies the resource

        :returns: twilio.rest.verify.v2.service.verification.VerificationContext
        :rtype: twilio.rest.verify.v2.service.verification.VerificationContext
        """
        return VerificationContext(self._version, service_sid=self._solution['service_sid'], sid=sid, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Verify.V2.VerificationList>'


class VerificationPage(Page):

    def __init__(self, version, response, solution):
        """
        Initialize the VerificationPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param service_sid: The SID of the Service that the resource is associated with

        :returns: twilio.rest.verify.v2.service.verification.VerificationPage
        :rtype: twilio.rest.verify.v2.service.verification.VerificationPage
        """
        super(VerificationPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of VerificationInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.verify.v2.service.verification.VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        return VerificationInstance(self._version, payload, service_sid=self._solution['service_sid'], )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Verify.V2.VerificationPage>'


class VerificationContext(InstanceContext):

    def __init__(self, version, service_sid, sid):
        """
        Initialize the VerificationContext

        :param Version version: Version that contains the resource
        :param service_sid: The SID of the verification Service to fetch the resource from
        :param sid: The unique string that identifies the resource

        :returns: twilio.rest.verify.v2.service.verification.VerificationContext
        :rtype: twilio.rest.verify.v2.service.verification.VerificationContext
        """
        super(VerificationContext, self).__init__(version)

        # Path Solution
        self._solution = {'service_sid': service_sid, 'sid': sid, }
        self._uri = '/Services/{service_sid}/Verifications/{sid}'.format(**self._solution)

    def update(self, status):
        """
        Update the VerificationInstance

        :param VerificationInstance.Status status: The new status of the resource

        :returns: The updated VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        data = values.of({'Status': status, })

        payload = self._version.update(method='POST', uri=self._uri, data=data, )

        return VerificationInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            sid=self._solution['sid'],
        )

    def fetch(self):
        """
        Fetch the VerificationInstance

        :returns: The fetched VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return VerificationInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            sid=self._solution['sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Verify.V2.VerificationContext {}>'.format(context)


class VerificationInstance(InstanceResource):

    class Channel(object):
        SMS = "sms"
        CALL = "call"
        EMAIL = "email"

    class Status(object):
        CANCELED = "canceled"
        APPROVED = "approved"

    def __init__(self, version, payload, service_sid, sid=None):
        """
        Initialize the VerificationInstance

        :returns: twilio.rest.verify.v2.service.verification.VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        super(VerificationInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'sid': payload.get('sid'),
            'service_sid': payload.get('service_sid'),
            'account_sid': payload.get('account_sid'),
            'to': payload.get('to'),
            'channel': payload.get('channel'),
            'status': payload.get('status'),
            'valid': payload.get('valid'),
            'lookup': payload.get('lookup'),
            'amount': payload.get('amount'),
            'payee': payload.get('payee'),
            'send_code_attempts': payload.get('send_code_attempts'),
            'date_created': deserialize.iso8601_datetime(payload.get('date_created')),
            'date_updated': deserialize.iso8601_datetime(payload.get('date_updated')),
            'url': payload.get('url'),
        }

        # Context
        self._context = None
        self._solution = {'service_sid': service_sid, 'sid': sid or self._properties['sid'], }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: VerificationContext for this VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationContext
        """
        if self._context is None:
            self._context = VerificationContext(
                self._version,
                service_sid=self._solution['service_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def sid(self):
        """
        :returns: The unique string that identifies the resource
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def service_sid(self):
        """
        :returns: The SID of the Service that the resource is associated with
        :rtype: unicode
        """
        return self._properties['service_sid']

    @property
    def account_sid(self):
        """
        :returns: The SID of the Account that created the resource
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def to(self):
        """
        :returns: The phone number or email being verified
        :rtype: unicode
        """
        return self._properties['to']

    @property
    def channel(self):
        """
        :returns: The verification method used.
        :rtype: VerificationInstance.Channel
        """
        return self._properties['channel']

    @property
    def status(self):
        """
        :returns: The status of the verification resource
        :rtype: unicode
        """
        return self._properties['status']

    @property
    def valid(self):
        """
        :returns: Whether the verification was successful
        :rtype: bool
        """
        return self._properties['valid']

    @property
    def lookup(self):
        """
        :returns: Information about the phone number being verified
        :rtype: dict
        """
        return self._properties['lookup']

    @property
    def amount(self):
        """
        :returns: The amount of the associated PSD2 compliant transaction.
        :rtype: unicode
        """
        return self._properties['amount']

    @property
    def payee(self):
        """
        :returns: The payee of the associated PSD2 compliant transaction
        :rtype: unicode
        """
        return self._properties['payee']

    @property
    def send_code_attempts(self):
        """
        :returns: An array of verification attempt objects.
        :rtype: dict
        """
        return self._properties['send_code_attempts']

    @property
    def date_created(self):
        """
        :returns: The RFC 2822 date and time in GMT when the resource was created
        :rtype: datetime
        """
        return self._properties['date_created']

    @property
    def date_updated(self):
        """
        :returns: The RFC 2822 date and time in GMT when the resource was last updated
        :rtype: datetime
        """
        return self._properties['date_updated']

    @property
    def url(self):
        """
        :returns: The absolute URL of the Verification resource
        :rtype: unicode
        """
        return self._properties['url']

    def update(self, status):
        """
        Update the VerificationInstance

        :param VerificationInstance.Status status: The new status of the resource

        :returns: The updated VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        return self._proxy.update(status, )

    def fetch(self):
        """
        Fetch the VerificationInstance

        :returns: The fetched VerificationInstance
        :rtype: twilio.rest.verify.v2.service.verification.VerificationInstance
        """
        return self._proxy.fetch()

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Verify.V2.VerificationInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import deserialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page
from twilio.rest.taskrouter.v1.workspace.worker.reservation import ReservationList
from twilio.rest.taskrouter.v1.workspace.worker.worker_channel import WorkerChannelList
from twilio.rest.taskrouter.v1.workspace.worker.worker_statistics import WorkerStatisticsList
from twilio.rest.taskrouter.v1.workspace.worker.workers_cumulative_statistics import WorkersCumulativeStatisticsList
from twilio.rest.taskrouter.v1.workspace.worker.workers_real_time_statistics import WorkersRealTimeStatisticsList
from twilio.rest.taskrouter.v1.workspace.worker.workers_statistics import WorkersStatisticsList


class WorkerList(ListResource):

    def __init__(self, version, workspace_sid):
        """
        Initialize the WorkerList

        :param Version version: Version that contains the resource
        :param workspace_sid: The SID of the Workspace that contains the Worker

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerList
        """
        super(WorkerList, self).__init__(version)

        # Path Solution
        self._solution = {'workspace_sid': workspace_sid, }
        self._uri = '/Workspaces/{workspace_sid}/Workers'.format(**self._solution)

        # Components
        self._statistics = None

    def stream(self, activity_name=values.unset, activity_sid=values.unset,
               available=values.unset, friendly_name=values.unset,
               target_workers_expression=values.unset, task_queue_name=values.unset,
               task_queue_sid=values.unset, limit=None, page_size=None):
        """
        Streams WorkerInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param unicode activity_name: The activity_name of the Worker resources to read
        :param unicode activity_sid: The activity_sid of the Worker resources to read
        :param unicode available: Whether to return Worker resources that are available or unavailable
        :param unicode friendly_name: The friendly_name of the Worker resources to read
        :param unicode target_workers_expression: Filter by Workers that would match an expression on a TaskQueue
        :param unicode task_queue_name: The friendly_name of the TaskQueue that the Workers to read are eligible for
        :param unicode task_queue_sid: The SID of the TaskQueue that the Workers to read are eligible for
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(
            activity_name=activity_name,
            activity_sid=activity_sid,
            available=available,
            friendly_name=friendly_name,
            target_workers_expression=target_workers_expression,
            task_queue_name=task_queue_name,
            task_queue_sid=task_queue_sid,
            page_size=limits['page_size'],
        )

        return self._version.stream(page, limits['limit'])

    def list(self, activity_name=values.unset, activity_sid=values.unset,
             available=values.unset, friendly_name=values.unset,
             target_workers_expression=values.unset, task_queue_name=values.unset,
             task_queue_sid=values.unset, limit=None, page_size=None):
        """
        Lists WorkerInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param unicode activity_name: The activity_name of the Worker resources to read
        :param unicode activity_sid: The activity_sid of the Worker resources to read
        :param unicode available: Whether to return Worker resources that are available or unavailable
        :param unicode friendly_name: The friendly_name of the Worker resources to read
        :param unicode target_workers_expression: Filter by Workers that would match an expression on a TaskQueue
        :param unicode task_queue_name: The friendly_name of the TaskQueue that the Workers to read are eligible for
        :param unicode task_queue_sid: The SID of the TaskQueue that the Workers to read are eligible for
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance]
        """
        return list(self.stream(
            activity_name=activity_name,
            activity_sid=activity_sid,
            available=available,
            friendly_name=friendly_name,
            target_workers_expression=target_workers_expression,
            task_queue_name=task_queue_name,
            task_queue_sid=task_queue_sid,
            limit=limit,
            page_size=page_size,
        ))

    def page(self, activity_name=values.unset, activity_sid=values.unset,
             available=values.unset, friendly_name=values.unset,
             target_workers_expression=values.unset, task_queue_name=values.unset,
             task_queue_sid=values.unset, page_token=values.unset,
             page_number=values.unset, page_size=values.unset):
        """
        Retrieve a single page of WorkerInstance records from the API.
        Request is executed immediately

        :param unicode activity_name: The activity_name of the Worker resources to read
        :param unicode activity_sid: The activity_sid of the Worker resources to read
        :param unicode available: Whether to return Worker resources that are available or unavailable
        :param unicode friendly_name: The friendly_name of the Worker resources to read
        :param unicode target_workers_expression: Filter by Workers that would match an expression on a TaskQueue
        :param unicode task_queue_name: The friendly_name of the TaskQueue that the Workers to read are eligible for
        :param unicode task_queue_sid: The SID of the TaskQueue that the Workers to read are eligible for
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerPage
        """
        data = values.of({
            'ActivityName': activity_name,
            'ActivitySid': activity_sid,
            'Available': available,
            'FriendlyName': friendly_name,
            'TargetWorkersExpression': target_workers_expression,
            'TaskQueueName': task_queue_name,
            'TaskQueueSid': task_queue_sid,
            'PageToken': page_token,
            'Page': page_number,
            'PageSize': page_size,
        })

        response = self._version.page(method='GET', uri=self._uri, params=data, )

        return WorkerPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of WorkerInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerPage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return WorkerPage(self._version, response, self._solution)

    def create(self, friendly_name, activity_sid=values.unset,
               attributes=values.unset):
        """
        Create the WorkerInstance

        :param unicode friendly_name: A string to describe the resource
        :param unicode activity_sid: The SID of a valid Activity that describes the new Worker's initial state
        :param unicode attributes: A valid JSON string that describes the new Worker

        :returns: The created WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        data = values.of({
            'FriendlyName': friendly_name,
            'ActivitySid': activity_sid,
            'Attributes': attributes,
        })

        payload = self._version.create(method='POST', uri=self._uri, data=data, )

        return WorkerInstance(self._version, payload, workspace_sid=self._solution['workspace_sid'], )

    @property
    def statistics(self):
        """
        Access the statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.workers_statistics.WorkersStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.workers_statistics.WorkersStatisticsList
        """
        if self._statistics is None:
            self._statistics = WorkersStatisticsList(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
            )
        return self._statistics

    def get(self, sid):
        """
        Constructs a WorkerContext

        :param sid: The SID of the resource to fetch

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        """
        return WorkerContext(self._version, workspace_sid=self._solution['workspace_sid'], sid=sid, )

    def __call__(self, sid):
        """
        Constructs a WorkerContext

        :param sid: The SID of the resource to fetch

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        """
        return WorkerContext(self._version, workspace_sid=self._solution['workspace_sid'], sid=sid, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Taskrouter.V1.WorkerList>'


class WorkerPage(Page):

    def __init__(self, version, response, solution):
        """
        Initialize the WorkerPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param workspace_sid: The SID of the Workspace that contains the Worker

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerPage
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerPage
        """
        super(WorkerPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of WorkerInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        return WorkerInstance(self._version, payload, workspace_sid=self._solution['workspace_sid'], )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Taskrouter.V1.WorkerPage>'


class WorkerContext(InstanceContext):

    def __init__(self, version, workspace_sid, sid):
        """
        Initialize the WorkerContext

        :param Version version: Version that contains the resource
        :param workspace_sid: The SID of the Workspace with the Worker to fetch
        :param sid: The SID of the resource to fetch

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        """
        super(WorkerContext, self).__init__(version)

        # Path Solution
        self._solution = {'workspace_sid': workspace_sid, 'sid': sid, }
        self._uri = '/Workspaces/{workspace_sid}/Workers/{sid}'.format(**self._solution)

        # Dependents
        self._real_time_statistics = None
        self._cumulative_statistics = None
        self._statistics = None
        self._reservations = None
        self._worker_channels = None

    def fetch(self):
        """
        Fetch the WorkerInstance

        :returns: The fetched WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return WorkerInstance(
            self._version,
            payload,
            workspace_sid=self._solution['workspace_sid'],
            sid=self._solution['sid'],
        )

    def update(self, activity_sid=values.unset, attributes=values.unset,
               friendly_name=values.unset,
               reject_pending_reservations=values.unset):
        """
        Update the WorkerInstance

        :param unicode activity_sid: The SID of the Activity that describes the Worker's initial state
        :param unicode attributes: The JSON string that describes the Worker
        :param unicode friendly_name: A string to describe the Worker
        :param bool reject_pending_reservations: Whether to reject pending reservations

        :returns: The updated WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        data = values.of({
            'ActivitySid': activity_sid,
            'Attributes': attributes,
            'FriendlyName': friendly_name,
            'RejectPendingReservations': reject_pending_reservations,
        })

        payload = self._version.update(method='POST', uri=self._uri, data=data, )

        return WorkerInstance(
            self._version,
            payload,
            workspace_sid=self._solution['workspace_sid'],
            sid=self._solution['sid'],
        )

    def delete(self):
        """
        Deletes the WorkerInstance

        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._version.delete(method='DELETE', uri=self._uri, )

    @property
    def real_time_statistics(self):
        """
        Access the real_time_statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.workers_real_time_statistics.WorkersRealTimeStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.workers_real_time_statistics.WorkersRealTimeStatisticsList
        """
        if self._real_time_statistics is None:
            self._real_time_statistics = WorkersRealTimeStatisticsList(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
            )
        return self._real_time_statistics

    @property
    def cumulative_statistics(self):
        """
        Access the cumulative_statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.workers_cumulative_statistics.WorkersCumulativeStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.workers_cumulative_statistics.WorkersCumulativeStatisticsList
        """
        if self._cumulative_statistics is None:
            self._cumulative_statistics = WorkersCumulativeStatisticsList(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
            )
        return self._cumulative_statistics

    @property
    def statistics(self):
        """
        Access the statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.worker_statistics.WorkerStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.worker_statistics.WorkerStatisticsList
        """
        if self._statistics is None:
            self._statistics = WorkerStatisticsList(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
                worker_sid=self._solution['sid'],
            )
        return self._statistics

    @property
    def reservations(self):
        """
        Access the reservations

        :returns: twilio.rest.taskrouter.v1.workspace.worker.reservation.ReservationList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.reservation.ReservationList
        """
        if self._reservations is None:
            self._reservations = ReservationList(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
                worker_sid=self._solution['sid'],
            )
        return self._reservations

    @property
    def worker_channels(self):
        """
        Access the worker_channels

        :returns: twilio.rest.taskrouter.v1.workspace.worker.worker_channel.WorkerChannelList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.worker_channel.WorkerChannelList
        """
        if self._worker_channels is None:
            self._worker_channels = WorkerChannelList(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
                worker_sid=self._solution['sid'],
            )
        return self._worker_channels

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Taskrouter.V1.WorkerContext {}>'.format(context)


class WorkerInstance(InstanceResource):

    def __init__(self, version, payload, workspace_sid, sid=None):
        """
        Initialize the WorkerInstance

        :returns: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        super(WorkerInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload.get('account_sid'),
            'activity_name': payload.get('activity_name'),
            'activity_sid': payload.get('activity_sid'),
            'attributes': payload.get('attributes'),
            'available': payload.get('available'),
            'date_created': deserialize.iso8601_datetime(payload.get('date_created')),
            'date_status_changed': deserialize.iso8601_datetime(payload.get('date_status_changed')),
            'date_updated': deserialize.iso8601_datetime(payload.get('date_updated')),
            'friendly_name': payload.get('friendly_name'),
            'sid': payload.get('sid'),
            'workspace_sid': payload.get('workspace_sid'),
            'url': payload.get('url'),
            'links': payload.get('links'),
        }

        # Context
        self._context = None
        self._solution = {'workspace_sid': workspace_sid, 'sid': sid or self._properties['sid'], }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: WorkerContext for this WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerContext
        """
        if self._context is None:
            self._context = WorkerContext(
                self._version,
                workspace_sid=self._solution['workspace_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: The SID of the Account that created the resource
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def activity_name(self):
        """
        :returns: The friendly_name of the Worker's current Activity
        :rtype: unicode
        """
        return self._properties['activity_name']

    @property
    def activity_sid(self):
        """
        :returns: The SID of the Worker's current Activity
        :rtype: unicode
        """
        return self._properties['activity_sid']

    @property
    def attributes(self):
        """
        :returns: The JSON string that describes the Worker
        :rtype: unicode
        """
        return self._properties['attributes']

    @property
    def available(self):
        """
        :returns: Whether the Worker is available to perform tasks
        :rtype: bool
        """
        return self._properties['available']

    @property
    def date_created(self):
        """
        :returns: The ISO 8601 date and time in GMT when the resource was created
        :rtype: datetime
        """
        return self._properties['date_created']

    @property
    def date_status_changed(self):
        """
        :returns: The date and time in GMT of the last change to the Worker's activity
        :rtype: datetime
        """
        return self._properties['date_status_changed']

    @property
    def date_updated(self):
        """
        :returns: The ISO 8601 date and time in GMT when the resource was last updated
        :rtype: datetime
        """
        return self._properties['date_updated']

    @property
    def friendly_name(self):
        """
        :returns: The string that you assigned to describe the resource
        :rtype: unicode
        """
        return self._properties['friendly_name']

    @property
    def sid(self):
        """
        :returns: The unique string that identifies the resource
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def workspace_sid(self):
        """
        :returns: The SID of the Workspace that contains the Worker
        :rtype: unicode
        """
        return self._properties['workspace_sid']

    @property
    def url(self):
        """
        :returns: The absolute URL of the Worker resource
        :rtype: unicode
        """
        return self._properties['url']

    @property
    def links(self):
        """
        :returns: The URLs of related resources
        :rtype: unicode
        """
        return self._properties['links']

    def fetch(self):
        """
        Fetch the WorkerInstance

        :returns: The fetched WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        return self._proxy.fetch()

    def update(self, activity_sid=values.unset, attributes=values.unset,
               friendly_name=values.unset,
               reject_pending_reservations=values.unset):
        """
        Update the WorkerInstance

        :param unicode activity_sid: The SID of the Activity that describes the Worker's initial state
        :param unicode attributes: The JSON string that describes the Worker
        :param unicode friendly_name: A string to describe the Worker
        :param bool reject_pending_reservations: Whether to reject pending reservations

        :returns: The updated WorkerInstance
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.WorkerInstance
        """
        return self._proxy.update(
            activity_sid=activity_sid,
            attributes=attributes,
            friendly_name=friendly_name,
            reject_pending_reservations=reject_pending_reservations,
        )

    def delete(self):
        """
        Deletes the WorkerInstance

        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._proxy.delete()

    @property
    def real_time_statistics(self):
        """
        Access the real_time_statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.workers_real_time_statistics.WorkersRealTimeStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.workers_real_time_statistics.WorkersRealTimeStatisticsList
        """
        return self._proxy.real_time_statistics

    @property
    def cumulative_statistics(self):
        """
        Access the cumulative_statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.workers_cumulative_statistics.WorkersCumulativeStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.workers_cumulative_statistics.WorkersCumulativeStatisticsList
        """
        return self._proxy.cumulative_statistics

    @property
    def statistics(self):
        """
        Access the statistics

        :returns: twilio.rest.taskrouter.v1.workspace.worker.worker_statistics.WorkerStatisticsList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.worker_statistics.WorkerStatisticsList
        """
        return self._proxy.statistics

    @property
    def reservations(self):
        """
        Access the reservations

        :returns: twilio.rest.taskrouter.v1.workspace.worker.reservation.ReservationList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.reservation.ReservationList
        """
        return self._proxy.reservations

    @property
    def worker_channels(self):
        """
        Access the worker_channels

        :returns: twilio.rest.taskrouter.v1.workspace.worker.worker_channel.WorkerChannelList
        :rtype: twilio.rest.taskrouter.v1.workspace.worker.worker_channel.WorkerChannelList
        """
        return self._proxy.worker_channels

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Taskrouter.V1.WorkerInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import deserialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class DeliveryReceiptList(ListResource):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, conversation_sid, message_sid):
        """
        Initialize the DeliveryReceiptList

        :param Version version: Version that contains the resource
        :param conversation_sid: The conversation_sid
        :param message_sid: The sid of the message the delivery receipt belongs to

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptList
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptList
        """
        super(DeliveryReceiptList, self).__init__(version)

        # Path Solution
        self._solution = {'conversation_sid': conversation_sid, 'message_sid': message_sid, }
        self._uri = '/Conversations/{conversation_sid}/Messages/{message_sid}/Receipts'.format(**self._solution)

    def stream(self, limit=None, page_size=None):
        """
        Streams DeliveryReceiptInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(page_size=limits['page_size'], )

        return self._version.stream(page, limits['limit'])

    def list(self, limit=None, page_size=None):
        """
        Lists DeliveryReceiptInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance]
        """
        return list(self.stream(limit=limit, page_size=page_size, ))

    def page(self, page_token=values.unset, page_number=values.unset,
             page_size=values.unset):
        """
        Retrieve a single page of DeliveryReceiptInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptPage
        """
        data = values.of({'PageToken': page_token, 'Page': page_number, 'PageSize': page_size, })

        response = self._version.page(method='GET', uri=self._uri, params=data, )

        return DeliveryReceiptPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of DeliveryReceiptInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptPage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return DeliveryReceiptPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a DeliveryReceiptContext

        :param sid: A 34 character string that uniquely identifies this resource.

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        """
        return DeliveryReceiptContext(
            self._version,
            conversation_sid=self._solution['conversation_sid'],
            message_sid=self._solution['message_sid'],
            sid=sid,
        )

    def __call__(self, sid):
        """
        Constructs a DeliveryReceiptContext

        :param sid: A 34 character string that uniquely identifies this resource.

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        """
        return DeliveryReceiptContext(
            self._version,
            conversation_sid=self._solution['conversation_sid'],
            message_sid=self._solution['message_sid'],
            sid=sid,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Conversations.V1.DeliveryReceiptList>'


class DeliveryReceiptPage(Page):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, response, solution):
        """
        Initialize the DeliveryReceiptPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param conversation_sid: The conversation_sid
        :param message_sid: The sid of the message the delivery receipt belongs to

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptPage
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptPage
        """
        super(DeliveryReceiptPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of DeliveryReceiptInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance
        """
        return DeliveryReceiptInstance(
            self._version,
            payload,
            conversation_sid=self._solution['conversation_sid'],
            message_sid=self._solution['message_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Conversations.V1.DeliveryReceiptPage>'


class DeliveryReceiptContext(InstanceContext):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    def __init__(self, version, conversation_sid, message_sid, sid):
        """
        Initialize the DeliveryReceiptContext

        :param Version version: Version that contains the resource
        :param conversation_sid: The unique id of the Conversation for this delivery receipt.
        :param message_sid: The sid of the message the delivery receipt belongs to
        :param sid: A 34 character string that uniquely identifies this resource.

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        """
        super(DeliveryReceiptContext, self).__init__(version)

        # Path Solution
        self._solution = {'conversation_sid': conversation_sid, 'message_sid': message_sid, 'sid': sid, }
        self._uri = '/Conversations/{conversation_sid}/Messages/{message_sid}/Receipts/{sid}'.format(**self._solution)

    def fetch(self):
        """
        Fetch the DeliveryReceiptInstance

        :returns: The fetched DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return DeliveryReceiptInstance(
            self._version,
            payload,
            conversation_sid=self._solution['conversation_sid'],
            message_sid=self._solution['message_sid'],
            sid=self._solution['sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Conversations.V1.DeliveryReceiptContext {}>'.format(context)


class DeliveryReceiptInstance(InstanceResource):
    """ PLEASE NOTE that this class contains beta products that are subject to
    change. Use them with caution. """

    class DeliveryStatus(object):
        READ = "read"
        FAILED = "failed"
        DELIVERED = "delivered"
        UNDELIVERED = "undelivered"
        SENT = "sent"

    def __init__(self, version, payload, conversation_sid, message_sid, sid=None):
        """
        Initialize the DeliveryReceiptInstance

        :returns: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance
        """
        super(DeliveryReceiptInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'sid': payload.get('sid'),
            'message_sid': payload.get('message_sid'),
            'conversation_sid': payload.get('conversation_sid'),
            'channel_message_sid': payload.get('channel_message_sid'),
            'participant_sid': payload.get('participant_sid'),
            'status': payload.get('status'),
            'error_code': deserialize.integer(payload.get('error_code')),
            'date_created': deserialize.iso8601_datetime(payload.get('date_created')),
            'date_updated': deserialize.iso8601_datetime(payload.get('date_updated')),
            'url': payload.get('url'),
        }

        # Context
        self._context = None
        self._solution = {
            'conversation_sid': conversation_sid,
            'message_sid': message_sid,
            'sid': sid or self._properties['sid'],
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: DeliveryReceiptContext for this DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptContext
        """
        if self._context is None:
            self._context = DeliveryReceiptContext(
                self._version,
                conversation_sid=self._solution['conversation_sid'],
                message_sid=self._solution['message_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def sid(self):
        """
        :returns: A 34 character string that uniquely identifies this resource.
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def message_sid(self):
        """
        :returns: The sid of the message the delivery receipt belongs to
        :rtype: unicode
        """
        return self._properties['message_sid']

    @property
    def conversation_sid(self):
        """
        :returns: The conversation_sid
        :rtype: unicode
        """
        return self._properties['conversation_sid']

    @property
    def channel_message_sid(self):
        """
        :returns: A messaging channel-specific identifier for the message delivered to participant
        :rtype: unicode
        """
        return self._properties['channel_message_sid']

    @property
    def participant_sid(self):
        """
        :returns: The unique id of the participant the delivery receipt belongs to.
        :rtype: unicode
        """
        return self._properties['participant_sid']

    @property
    def status(self):
        """
        :returns: The message delivery status
        :rtype: DeliveryReceiptInstance.DeliveryStatus
        """
        return self._properties['status']

    @property
    def error_code(self):
        """
        :returns: The message `delivery error code <https://www.twilio.com/docs/sms/api/message-resource#delivery-related-errors>`_ for a `failed` status
        :rtype: unicode
        """
        return self._properties['error_code']

    @property
    def date_created(self):
        """
        :returns: The date that this resource was created.
        :rtype: datetime
        """
        return self._properties['date_created']

    @property
    def date_updated(self):
        """
        :returns: The date that this resource was last updated.
        :rtype: datetime
        """
        return self._properties['date_updated']

    @property
    def url(self):
        """
        :returns: An absolute URL for this delivery receipt.
        :rtype: unicode
        """
        return self._properties['url']

    def fetch(self):
        """
        Fetch the DeliveryReceiptInstance

        :returns: The fetched DeliveryReceiptInstance
        :rtype: twilio.rest.conversations.v1.conversation.message.delivery_receipt.DeliveryReceiptInstance
        """
        return self._proxy.fetch()

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Conversations.V1.DeliveryReceiptInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base.version import Version
from twilio.rest.preview.marketplace.available_add_on import AvailableAddOnList
from twilio.rest.preview.marketplace.installed_add_on import InstalledAddOnList


class Marketplace(Version):

    def __init__(self, domain):
        """
        Initialize the Marketplace version of Preview

        :returns: Marketplace version of Preview
        :rtype: twilio.rest.preview.marketplace.Marketplace.Marketplace
        """
        super(Marketplace, self).__init__(domain)
        self.version = 'marketplace'
        self._available_add_ons = None
        self._installed_add_ons = None

    @property
    def available_add_ons(self):
        """
        :rtype: twilio.rest.preview.marketplace.available_add_on.AvailableAddOnList
        """
        if self._available_add_ons is None:
            self._available_add_ons = AvailableAddOnList(self)
        return self._available_add_ons

    @property
    def installed_add_ons(self):
        """
        :rtype: twilio.rest.preview.marketplace.installed_add_on.InstalledAddOnList
        """
        if self._installed_add_ons is None:
            self._installed_add_ons = InstalledAddOnList(self)
        return self._installed_add_ons

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.Marketplace>'



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import deserialize
from twilio.base import serialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page


class LogList(ListResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, service_sid, environment_sid):
        """
        Initialize the LogList

        :param Version version: Version that contains the resource
        :param service_sid: The SID of the Service that the Log resource is associated with
        :param environment_sid: The SID of the environment in which the log occurred

        :returns: twilio.rest.serverless.v1.service.environment.log.LogList
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogList
        """
        super(LogList, self).__init__(version)

        # Path Solution
        self._solution = {'service_sid': service_sid, 'environment_sid': environment_sid, }
        self._uri = '/Services/{service_sid}/Environments/{environment_sid}/Logs'.format(**self._solution)

    def stream(self, function_sid=values.unset, start_date=values.unset,
               end_date=values.unset, limit=None, page_size=None):
        """
        Streams LogInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param unicode function_sid: The SID of the function whose invocation produced the Log resources to read
        :param datetime start_date: The date and time after which the Log resources must have been created.
        :param datetime end_date: The date and time before which the Log resource must have been created.
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.serverless.v1.service.environment.log.LogInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(
            function_sid=function_sid,
            start_date=start_date,
            end_date=end_date,
            page_size=limits['page_size'],
        )

        return self._version.stream(page, limits['limit'])

    def list(self, function_sid=values.unset, start_date=values.unset,
             end_date=values.unset, limit=None, page_size=None):
        """
        Lists LogInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param unicode function_sid: The SID of the function whose invocation produced the Log resources to read
        :param datetime start_date: The date and time after which the Log resources must have been created.
        :param datetime end_date: The date and time before which the Log resource must have been created.
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.serverless.v1.service.environment.log.LogInstance]
        """
        return list(self.stream(
            function_sid=function_sid,
            start_date=start_date,
            end_date=end_date,
            limit=limit,
            page_size=page_size,
        ))

    def page(self, function_sid=values.unset, start_date=values.unset,
             end_date=values.unset, page_token=values.unset,
             page_number=values.unset, page_size=values.unset):
        """
        Retrieve a single page of LogInstance records from the API.
        Request is executed immediately

        :param unicode function_sid: The SID of the function whose invocation produced the Log resources to read
        :param datetime start_date: The date and time after which the Log resources must have been created.
        :param datetime end_date: The date and time before which the Log resource must have been created.
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogPage
        """
        data = values.of({
            'FunctionSid': function_sid,
            'StartDate': serialize.iso8601_datetime(start_date),
            'EndDate': serialize.iso8601_datetime(end_date),
            'PageToken': page_token,
            'Page': page_number,
            'PageSize': page_size,
        })

        response = self._version.page(method='GET', uri=self._uri, params=data, )

        return LogPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of LogInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogPage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return LogPage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a LogContext

        :param sid: The SID that identifies the Log resource to fetch

        :returns: twilio.rest.serverless.v1.service.environment.log.LogContext
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogContext
        """
        return LogContext(
            self._version,
            service_sid=self._solution['service_sid'],
            environment_sid=self._solution['environment_sid'],
            sid=sid,
        )

    def __call__(self, sid):
        """
        Constructs a LogContext

        :param sid: The SID that identifies the Log resource to fetch

        :returns: twilio.rest.serverless.v1.service.environment.log.LogContext
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogContext
        """
        return LogContext(
            self._version,
            service_sid=self._solution['service_sid'],
            environment_sid=self._solution['environment_sid'],
            sid=sid,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Serverless.V1.LogList>'


class LogPage(Page):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, response, solution):
        """
        Initialize the LogPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param service_sid: The SID of the Service that the Log resource is associated with
        :param environment_sid: The SID of the environment in which the log occurred

        :returns: twilio.rest.serverless.v1.service.environment.log.LogPage
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogPage
        """
        super(LogPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of LogInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.serverless.v1.service.environment.log.LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogInstance
        """
        return LogInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            environment_sid=self._solution['environment_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Serverless.V1.LogPage>'


class LogContext(InstanceContext):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, service_sid, environment_sid, sid):
        """
        Initialize the LogContext

        :param Version version: Version that contains the resource
        :param service_sid: The SID of the Service to fetch the Log resource from
        :param environment_sid: The SID of the environment with the Log resource to fetch
        :param sid: The SID that identifies the Log resource to fetch

        :returns: twilio.rest.serverless.v1.service.environment.log.LogContext
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogContext
        """
        super(LogContext, self).__init__(version)

        # Path Solution
        self._solution = {'service_sid': service_sid, 'environment_sid': environment_sid, 'sid': sid, }
        self._uri = '/Services/{service_sid}/Environments/{environment_sid}/Logs/{sid}'.format(**self._solution)

    def fetch(self):
        """
        Fetch the LogInstance

        :returns: The fetched LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return LogInstance(
            self._version,
            payload,
            service_sid=self._solution['service_sid'],
            environment_sid=self._solution['environment_sid'],
            sid=self._solution['sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Serverless.V1.LogContext {}>'.format(context)


class LogInstance(InstanceResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    class Level(object):
        INFO = "info"
        WARN = "warn"
        ERROR = "error"

    def __init__(self, version, payload, service_sid, environment_sid, sid=None):
        """
        Initialize the LogInstance

        :returns: twilio.rest.serverless.v1.service.environment.log.LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogInstance
        """
        super(LogInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'sid': payload.get('sid'),
            'account_sid': payload.get('account_sid'),
            'service_sid': payload.get('service_sid'),
            'environment_sid': payload.get('environment_sid'),
            'build_sid': payload.get('build_sid'),
            'deployment_sid': payload.get('deployment_sid'),
            'function_sid': payload.get('function_sid'),
            'request_sid': payload.get('request_sid'),
            'level': payload.get('level'),
            'message': payload.get('message'),
            'date_created': deserialize.iso8601_datetime(payload.get('date_created')),
            'url': payload.get('url'),
        }

        # Context
        self._context = None
        self._solution = {
            'service_sid': service_sid,
            'environment_sid': environment_sid,
            'sid': sid or self._properties['sid'],
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: LogContext for this LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogContext
        """
        if self._context is None:
            self._context = LogContext(
                self._version,
                service_sid=self._solution['service_sid'],
                environment_sid=self._solution['environment_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def sid(self):
        """
        :returns: The unique string that identifies the Log resource
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def account_sid(self):
        """
        :returns: The SID of the Account that created the Log resource
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def service_sid(self):
        """
        :returns: The SID of the Service that the Log resource is associated with
        :rtype: unicode
        """
        return self._properties['service_sid']

    @property
    def environment_sid(self):
        """
        :returns: The SID of the environment in which the log occurred
        :rtype: unicode
        """
        return self._properties['environment_sid']

    @property
    def build_sid(self):
        """
        :returns: The SID of the build that corresponds to the log
        :rtype: unicode
        """
        return self._properties['build_sid']

    @property
    def deployment_sid(self):
        """
        :returns: The SID of the deployment that corresponds to the log
        :rtype: unicode
        """
        return self._properties['deployment_sid']

    @property
    def function_sid(self):
        """
        :returns: The SID of the function whose invocation produced the log
        :rtype: unicode
        """
        return self._properties['function_sid']

    @property
    def request_sid(self):
        """
        :returns: The SID of the request associated with the log
        :rtype: unicode
        """
        return self._properties['request_sid']

    @property
    def level(self):
        """
        :returns: The log level
        :rtype: LogInstance.Level
        """
        return self._properties['level']

    @property
    def message(self):
        """
        :returns: The log message
        :rtype: unicode
        """
        return self._properties['message']

    @property
    def date_created(self):
        """
        :returns: The ISO 8601 date and time in GMT when the Log resource was created
        :rtype: datetime
        """
        return self._properties['date_created']

    @property
    def url(self):
        """
        :returns: The absolute URL of the Log resource
        :rtype: unicode
        """
        return self._properties['url']

    def fetch(self):
        """
        Fetch the LogInstance

        :returns: The fetched LogInstance
        :rtype: twilio.rest.serverless.v1.service.environment.log.LogInstance
        """
        return self._proxy.fetch()

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Serverless.V1.LogInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page
from twilio.rest.preview.trusted_comms.business.brand.branded_channel.channel import ChannelList


class BrandedChannelList(ListResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, business_sid, brand_sid):
        """
        Initialize the BrandedChannelList

        :param Version version: Version that contains the resource
        :param business_sid: Business Sid.
        :param brand_sid: Brand Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelList
        """
        super(BrandedChannelList, self).__init__(version)

        # Path Solution
        self._solution = {'business_sid': business_sid, 'brand_sid': brand_sid, }

    def get(self, sid):
        """
        Constructs a BrandedChannelContext

        :param sid: Branded Channel Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        """
        return BrandedChannelContext(
            self._version,
            business_sid=self._solution['business_sid'],
            brand_sid=self._solution['brand_sid'],
            sid=sid,
        )

    def __call__(self, sid):
        """
        Constructs a BrandedChannelContext

        :param sid: Branded Channel Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        """
        return BrandedChannelContext(
            self._version,
            business_sid=self._solution['business_sid'],
            brand_sid=self._solution['brand_sid'],
            sid=sid,
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.TrustedComms.BrandedChannelList>'


class BrandedChannelPage(Page):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, response, solution):
        """
        Initialize the BrandedChannelPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API
        :param business_sid: Business Sid.
        :param brand_sid: Brand Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelPage
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelPage
        """
        super(BrandedChannelPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of BrandedChannelInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelInstance
        """
        return BrandedChannelInstance(
            self._version,
            payload,
            business_sid=self._solution['business_sid'],
            brand_sid=self._solution['brand_sid'],
        )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.TrustedComms.BrandedChannelPage>'


class BrandedChannelContext(InstanceContext):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, business_sid, brand_sid, sid):
        """
        Initialize the BrandedChannelContext

        :param Version version: Version that contains the resource
        :param business_sid: Business Sid.
        :param brand_sid: Brand Sid.
        :param sid: Branded Channel Sid.

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        """
        super(BrandedChannelContext, self).__init__(version)

        # Path Solution
        self._solution = {'business_sid': business_sid, 'brand_sid': brand_sid, 'sid': sid, }
        self._uri = '/Businesses/{business_sid}/Brands/{brand_sid}/BrandedChannels/{sid}'.format(**self._solution)

        # Dependents
        self._channels = None

    def fetch(self):
        """
        Fetch the BrandedChannelInstance

        :returns: The fetched BrandedChannelInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return BrandedChannelInstance(
            self._version,
            payload,
            business_sid=self._solution['business_sid'],
            brand_sid=self._solution['brand_sid'],
            sid=self._solution['sid'],
        )

    @property
    def channels(self):
        """
        Access the channels

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.channel.ChannelList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.channel.ChannelList
        """
        if self._channels is None:
            self._channels = ChannelList(
                self._version,
                business_sid=self._solution['business_sid'],
                brand_sid=self._solution['brand_sid'],
                branded_channel_sid=self._solution['sid'],
            )
        return self._channels

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.TrustedComms.BrandedChannelContext {}>'.format(context)


class BrandedChannelInstance(InstanceResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, payload, business_sid, brand_sid, sid=None):
        """
        Initialize the BrandedChannelInstance

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelInstance
        """
        super(BrandedChannelInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload.get('account_sid'),
            'business_sid': payload.get('business_sid'),
            'brand_sid': payload.get('brand_sid'),
            'sid': payload.get('sid'),
            'links': payload.get('links'),
            'url': payload.get('url'),
        }

        # Context
        self._context = None
        self._solution = {
            'business_sid': business_sid,
            'brand_sid': brand_sid,
            'sid': sid or self._properties['sid'],
        }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: BrandedChannelContext for this BrandedChannelInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelContext
        """
        if self._context is None:
            self._context = BrandedChannelContext(
                self._version,
                business_sid=self._solution['business_sid'],
                brand_sid=self._solution['brand_sid'],
                sid=self._solution['sid'],
            )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: Account Sid.
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def business_sid(self):
        """
        :returns: Business Sid.
        :rtype: unicode
        """
        return self._properties['business_sid']

    @property
    def brand_sid(self):
        """
        :returns: Brand Sid.
        :rtype: unicode
        """
        return self._properties['brand_sid']

    @property
    def sid(self):
        """
        :returns: Branded Channel Sid.
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def links(self):
        """
        :returns: Nested resource URLs.
        :rtype: unicode
        """
        return self._properties['links']

    @property
    def url(self):
        """
        :returns: The URL of this resource.
        :rtype: unicode
        """
        return self._properties['url']

    def fetch(self):
        """
        Fetch the BrandedChannelInstance

        :returns: The fetched BrandedChannelInstance
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.BrandedChannelInstance
        """
        return self._proxy.fetch()

    @property
    def channels(self):
        """
        Access the channels

        :returns: twilio.rest.preview.trusted_comms.business.brand.branded_channel.channel.ChannelList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.branded_channel.channel.ChannelList
        """
        return self._proxy.channels

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.TrustedComms.BrandedChannelInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page
from twilio.rest.preview.trusted_comms.business.brand import BrandList
from twilio.rest.preview.trusted_comms.business.insights import InsightsList


class BusinessList(ListResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version):
        """
        Initialize the BusinessList

        :param Version version: Version that contains the resource

        :returns: twilio.rest.preview.trusted_comms.business.BusinessList
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessList
        """
        super(BusinessList, self).__init__(version)

        # Path Solution
        self._solution = {}

    def get(self, sid):
        """
        Constructs a BusinessContext

        :param sid: A string that uniquely identifies this Business.

        :returns: twilio.rest.preview.trusted_comms.business.BusinessContext
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessContext
        """
        return BusinessContext(self._version, sid=sid, )

    def __call__(self, sid):
        """
        Constructs a BusinessContext

        :param sid: A string that uniquely identifies this Business.

        :returns: twilio.rest.preview.trusted_comms.business.BusinessContext
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessContext
        """
        return BusinessContext(self._version, sid=sid, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.TrustedComms.BusinessList>'


class BusinessPage(Page):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, response, solution):
        """
        Initialize the BusinessPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API

        :returns: twilio.rest.preview.trusted_comms.business.BusinessPage
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessPage
        """
        super(BusinessPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of BusinessInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.preview.trusted_comms.business.BusinessInstance
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessInstance
        """
        return BusinessInstance(self._version, payload, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.TrustedComms.BusinessPage>'


class BusinessContext(InstanceContext):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, sid):
        """
        Initialize the BusinessContext

        :param Version version: Version that contains the resource
        :param sid: A string that uniquely identifies this Business.

        :returns: twilio.rest.preview.trusted_comms.business.BusinessContext
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessContext
        """
        super(BusinessContext, self).__init__(version)

        # Path Solution
        self._solution = {'sid': sid, }
        self._uri = '/Businesses/{sid}'.format(**self._solution)

        # Dependents
        self._brands = None
        self._insights = None

    def fetch(self):
        """
        Fetch the BusinessInstance

        :returns: The fetched BusinessInstance
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return BusinessInstance(self._version, payload, sid=self._solution['sid'], )

    @property
    def brands(self):
        """
        Access the brands

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandList
        """
        if self._brands is None:
            self._brands = BrandList(self._version, business_sid=self._solution['sid'], )
        return self._brands

    @property
    def insights(self):
        """
        Access the insights

        :returns: twilio.rest.preview.trusted_comms.business.insights.InsightsList
        :rtype: twilio.rest.preview.trusted_comms.business.insights.InsightsList
        """
        if self._insights is None:
            self._insights = InsightsList(self._version, business_sid=self._solution['sid'], )
        return self._insights

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.TrustedComms.BusinessContext {}>'.format(context)


class BusinessInstance(InstanceResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, payload, sid=None):
        """
        Initialize the BusinessInstance

        :returns: twilio.rest.preview.trusted_comms.business.BusinessInstance
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessInstance
        """
        super(BusinessInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'account_sid': payload.get('account_sid'),
            'sid': payload.get('sid'),
            'url': payload.get('url'),
            'links': payload.get('links'),
        }

        # Context
        self._context = None
        self._solution = {'sid': sid or self._properties['sid'], }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: BusinessContext for this BusinessInstance
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessContext
        """
        if self._context is None:
            self._context = BusinessContext(self._version, sid=self._solution['sid'], )
        return self._context

    @property
    def account_sid(self):
        """
        :returns: Account Sid.
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def sid(self):
        """
        :returns: A string that uniquely identifies this Business.
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def url(self):
        """
        :returns: The URL of this resource.
        :rtype: unicode
        """
        return self._properties['url']

    @property
    def links(self):
        """
        :returns: Nested resource URLs.
        :rtype: unicode
        """
        return self._properties['links']

    def fetch(self):
        """
        Fetch the BusinessInstance

        :returns: The fetched BusinessInstance
        :rtype: twilio.rest.preview.trusted_comms.business.BusinessInstance
        """
        return self._proxy.fetch()

    @property
    def brands(self):
        """
        Access the brands

        :returns: twilio.rest.preview.trusted_comms.business.brand.BrandList
        :rtype: twilio.rest.preview.trusted_comms.business.brand.BrandList
        """
        return self._proxy.brands

    @property
    def insights(self):
        """
        Access the insights

        :returns: twilio.rest.preview.trusted_comms.business.insights.InsightsList
        :rtype: twilio.rest.preview.trusted_comms.business.insights.InsightsList
        """
        return self._proxy.insights

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.TrustedComms.BusinessInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import deserialize
from twilio.base import serialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page
from twilio.rest.preview.hosted_numbers.authorization_document.dependent_hosted_number_order import DependentHostedNumberOrderList


class AuthorizationDocumentList(ListResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version):
        """
        Initialize the AuthorizationDocumentList

        :param Version version: Version that contains the resource

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentList
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentList
        """
        super(AuthorizationDocumentList, self).__init__(version)

        # Path Solution
        self._solution = {}
        self._uri = '/AuthorizationDocuments'.format(**self._solution)

    def stream(self, email=values.unset, status=values.unset, limit=None,
               page_size=None):
        """
        Streams AuthorizationDocumentInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param unicode email: Email.
        :param AuthorizationDocumentInstance.Status status: The Status of this AuthorizationDocument.
        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(email=email, status=status, page_size=limits['page_size'], )

        return self._version.stream(page, limits['limit'])

    def list(self, email=values.unset, status=values.unset, limit=None,
             page_size=None):
        """
        Lists AuthorizationDocumentInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param unicode email: Email.
        :param AuthorizationDocumentInstance.Status status: The Status of this AuthorizationDocument.
        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance]
        """
        return list(self.stream(email=email, status=status, limit=limit, page_size=page_size, ))

    def page(self, email=values.unset, status=values.unset, page_token=values.unset,
             page_number=values.unset, page_size=values.unset):
        """
        Retrieve a single page of AuthorizationDocumentInstance records from the API.
        Request is executed immediately

        :param unicode email: Email.
        :param AuthorizationDocumentInstance.Status status: The Status of this AuthorizationDocument.
        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentPage
        """
        data = values.of({
            'Email': email,
            'Status': status,
            'PageToken': page_token,
            'Page': page_number,
            'PageSize': page_size,
        })

        response = self._version.page(method='GET', uri=self._uri, params=data, )

        return AuthorizationDocumentPage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of AuthorizationDocumentInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentPage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return AuthorizationDocumentPage(self._version, response, self._solution)

    def create(self, hosted_number_order_sids, address_sid, email, contact_title,
               contact_phone_number, cc_emails=values.unset):
        """
        Create the AuthorizationDocumentInstance

        :param unicode hosted_number_order_sids: A list of HostedNumberOrder sids.
        :param unicode address_sid: Address sid.
        :param unicode email: Email.
        :param unicode contact_title: Title of signee of this Authorization Document.
        :param unicode contact_phone_number: Authorization Document's signee's phone number.
        :param unicode cc_emails: A list of emails.

        :returns: The created AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        data = values.of({
            'HostedNumberOrderSids': serialize.map(hosted_number_order_sids, lambda e: e),
            'AddressSid': address_sid,
            'Email': email,
            'ContactTitle': contact_title,
            'ContactPhoneNumber': contact_phone_number,
            'CcEmails': serialize.map(cc_emails, lambda e: e),
        })

        payload = self._version.create(method='POST', uri=self._uri, data=data, )

        return AuthorizationDocumentInstance(self._version, payload, )

    def get(self, sid):
        """
        Constructs a AuthorizationDocumentContext

        :param sid: AuthorizationDocument sid.

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        """
        return AuthorizationDocumentContext(self._version, sid=sid, )

    def __call__(self, sid):
        """
        Constructs a AuthorizationDocumentContext

        :param sid: AuthorizationDocument sid.

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        """
        return AuthorizationDocumentContext(self._version, sid=sid, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.HostedNumbers.AuthorizationDocumentList>'


class AuthorizationDocumentPage(Page):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, response, solution):
        """
        Initialize the AuthorizationDocumentPage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentPage
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentPage
        """
        super(AuthorizationDocumentPage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of AuthorizationDocumentInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        return AuthorizationDocumentInstance(self._version, payload, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Preview.HostedNumbers.AuthorizationDocumentPage>'


class AuthorizationDocumentContext(InstanceContext):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    def __init__(self, version, sid):
        """
        Initialize the AuthorizationDocumentContext

        :param Version version: Version that contains the resource
        :param sid: AuthorizationDocument sid.

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        """
        super(AuthorizationDocumentContext, self).__init__(version)

        # Path Solution
        self._solution = {'sid': sid, }
        self._uri = '/AuthorizationDocuments/{sid}'.format(**self._solution)

        # Dependents
        self._dependent_hosted_number_orders = None

    def fetch(self):
        """
        Fetch the AuthorizationDocumentInstance

        :returns: The fetched AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return AuthorizationDocumentInstance(self._version, payload, sid=self._solution['sid'], )

    def update(self, hosted_number_order_sids=values.unset,
               address_sid=values.unset, email=values.unset, cc_emails=values.unset,
               status=values.unset, contact_title=values.unset,
               contact_phone_number=values.unset):
        """
        Update the AuthorizationDocumentInstance

        :param unicode hosted_number_order_sids: A list of HostedNumberOrder sids.
        :param unicode address_sid: Address sid.
        :param unicode email: Email.
        :param unicode cc_emails: A list of emails.
        :param AuthorizationDocumentInstance.Status status: The Status of this AuthorizationDocument.
        :param unicode contact_title: Title of signee of this Authorization Document.
        :param unicode contact_phone_number: Authorization Document's signee's phone number.

        :returns: The updated AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        data = values.of({
            'HostedNumberOrderSids': serialize.map(hosted_number_order_sids, lambda e: e),
            'AddressSid': address_sid,
            'Email': email,
            'CcEmails': serialize.map(cc_emails, lambda e: e),
            'Status': status,
            'ContactTitle': contact_title,
            'ContactPhoneNumber': contact_phone_number,
        })

        payload = self._version.update(method='POST', uri=self._uri, data=data, )

        return AuthorizationDocumentInstance(self._version, payload, sid=self._solution['sid'], )

    @property
    def dependent_hosted_number_orders(self):
        """
        Access the dependent_hosted_number_orders

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.dependent_hosted_number_order.DependentHostedNumberOrderList
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.dependent_hosted_number_order.DependentHostedNumberOrderList
        """
        if self._dependent_hosted_number_orders is None:
            self._dependent_hosted_number_orders = DependentHostedNumberOrderList(
                self._version,
                signing_document_sid=self._solution['sid'],
            )
        return self._dependent_hosted_number_orders

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.HostedNumbers.AuthorizationDocumentContext {}>'.format(context)


class AuthorizationDocumentInstance(InstanceResource):
    """ PLEASE NOTE that this class contains preview products that are subject
    to change. Use them with caution. If you currently do not have developer
    preview access, please contact help@twilio.com. """

    class Status(object):
        OPENED = "opened"
        SIGNING = "signing"
        SIGNED = "signed"
        CANCELED = "canceled"
        FAILED = "failed"

    def __init__(self, version, payload, sid=None):
        """
        Initialize the AuthorizationDocumentInstance

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        super(AuthorizationDocumentInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'sid': payload.get('sid'),
            'address_sid': payload.get('address_sid'),
            'status': payload.get('status'),
            'email': payload.get('email'),
            'cc_emails': payload.get('cc_emails'),
            'date_created': deserialize.iso8601_datetime(payload.get('date_created')),
            'date_updated': deserialize.iso8601_datetime(payload.get('date_updated')),
            'url': payload.get('url'),
            'links': payload.get('links'),
        }

        # Context
        self._context = None
        self._solution = {'sid': sid or self._properties['sid'], }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: AuthorizationDocumentContext for this AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentContext
        """
        if self._context is None:
            self._context = AuthorizationDocumentContext(self._version, sid=self._solution['sid'], )
        return self._context

    @property
    def sid(self):
        """
        :returns: AuthorizationDocument sid.
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def address_sid(self):
        """
        :returns: Address sid.
        :rtype: unicode
        """
        return self._properties['address_sid']

    @property
    def status(self):
        """
        :returns: The Status of this AuthorizationDocument.
        :rtype: AuthorizationDocumentInstance.Status
        """
        return self._properties['status']

    @property
    def email(self):
        """
        :returns: Email.
        :rtype: unicode
        """
        return self._properties['email']

    @property
    def cc_emails(self):
        """
        :returns: A list of emails.
        :rtype: unicode
        """
        return self._properties['cc_emails']

    @property
    def date_created(self):
        """
        :returns: The date this AuthorizationDocument was created.
        :rtype: datetime
        """
        return self._properties['date_created']

    @property
    def date_updated(self):
        """
        :returns: The date this AuthorizationDocument was updated.
        :rtype: datetime
        """
        return self._properties['date_updated']

    @property
    def url(self):
        """
        :returns: The url
        :rtype: unicode
        """
        return self._properties['url']

    @property
    def links(self):
        """
        :returns: The links
        :rtype: unicode
        """
        return self._properties['links']

    def fetch(self):
        """
        Fetch the AuthorizationDocumentInstance

        :returns: The fetched AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        return self._proxy.fetch()

    def update(self, hosted_number_order_sids=values.unset,
               address_sid=values.unset, email=values.unset, cc_emails=values.unset,
               status=values.unset, contact_title=values.unset,
               contact_phone_number=values.unset):
        """
        Update the AuthorizationDocumentInstance

        :param unicode hosted_number_order_sids: A list of HostedNumberOrder sids.
        :param unicode address_sid: Address sid.
        :param unicode email: Email.
        :param unicode cc_emails: A list of emails.
        :param AuthorizationDocumentInstance.Status status: The Status of this AuthorizationDocument.
        :param unicode contact_title: Title of signee of this Authorization Document.
        :param unicode contact_phone_number: Authorization Document's signee's phone number.

        :returns: The updated AuthorizationDocumentInstance
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.AuthorizationDocumentInstance
        """
        return self._proxy.update(
            hosted_number_order_sids=hosted_number_order_sids,
            address_sid=address_sid,
            email=email,
            cc_emails=cc_emails,
            status=status,
            contact_title=contact_title,
            contact_phone_number=contact_phone_number,
        )

    @property
    def dependent_hosted_number_orders(self):
        """
        Access the dependent_hosted_number_orders

        :returns: twilio.rest.preview.hosted_numbers.authorization_document.dependent_hosted_number_order.DependentHostedNumberOrderList
        :rtype: twilio.rest.preview.hosted_numbers.authorization_document.dependent_hosted_number_order.DependentHostedNumberOrderList
        """
        return self._proxy.dependent_hosted_number_orders

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Preview.HostedNumbers.AuthorizationDocumentInstance {}>'.format(context)



# coding=utf-8
r"""
This code was generated by
\ / _    _  _|   _  _
 | (_)\/(_)(_|\/| |(/_  v1.0.0
      /       /
"""

from twilio.base import deserialize
from twilio.base import serialize
from twilio.base import values
from twilio.base.instance_context import InstanceContext
from twilio.base.instance_resource import InstanceResource
from twilio.base.list_resource import ListResource
from twilio.base.page import Page
from twilio.rest.chat.v1.service.channel import ChannelList
from twilio.rest.chat.v1.service.role import RoleList
from twilio.rest.chat.v1.service.user import UserList


class ServiceList(ListResource):

    def __init__(self, version):
        """
        Initialize the ServiceList

        :param Version version: Version that contains the resource

        :returns: twilio.rest.chat.v1.service.ServiceList
        :rtype: twilio.rest.chat.v1.service.ServiceList
        """
        super(ServiceList, self).__init__(version)

        # Path Solution
        self._solution = {}
        self._uri = '/Services'.format(**self._solution)

    def create(self, friendly_name):
        """
        Create the ServiceInstance

        :param unicode friendly_name: A string to describe the resource

        :returns: The created ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        data = values.of({'FriendlyName': friendly_name, })

        payload = self._version.create(method='POST', uri=self._uri, data=data, )

        return ServiceInstance(self._version, payload, )

    def stream(self, limit=None, page_size=None):
        """
        Streams ServiceInstance records from the API as a generator stream.
        This operation lazily loads records as efficiently as possible until the limit
        is reached.
        The results are returned as a generator, so this operation is memory efficient.

        :param int limit: Upper limit for the number of records to return. stream()
                          guarantees to never return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, stream() will attempt to read the
                              limit with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.chat.v1.service.ServiceInstance]
        """
        limits = self._version.read_limits(limit, page_size)

        page = self.page(page_size=limits['page_size'], )

        return self._version.stream(page, limits['limit'])

    def list(self, limit=None, page_size=None):
        """
        Lists ServiceInstance records from the API as a list.
        Unlike stream(), this operation is eager and will load `limit` records into
        memory before returning.

        :param int limit: Upper limit for the number of records to return. list() guarantees
                          never to return more than limit.  Default is no limit
        :param int page_size: Number of records to fetch per request, when not set will use
                              the default value of 50 records.  If no page_size is defined
                              but a limit is defined, list() will attempt to read the limit
                              with the most efficient page size, i.e. min(limit, 1000)

        :returns: Generator that will yield up to limit results
        :rtype: list[twilio.rest.chat.v1.service.ServiceInstance]
        """
        return list(self.stream(limit=limit, page_size=page_size, ))

    def page(self, page_token=values.unset, page_number=values.unset,
             page_size=values.unset):
        """
        Retrieve a single page of ServiceInstance records from the API.
        Request is executed immediately

        :param str page_token: PageToken provided by the API
        :param int page_number: Page Number, this value is simply for client state
        :param int page_size: Number of records to return, defaults to 50

        :returns: Page of ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServicePage
        """
        data = values.of({'PageToken': page_token, 'Page': page_number, 'PageSize': page_size, })

        response = self._version.page(method='GET', uri=self._uri, params=data, )

        return ServicePage(self._version, response, self._solution)

    def get_page(self, target_url):
        """
        Retrieve a specific page of ServiceInstance records from the API.
        Request is executed immediately

        :param str target_url: API-generated URL for the requested results page

        :returns: Page of ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServicePage
        """
        response = self._version.domain.twilio.request(
            'GET',
            target_url,
        )

        return ServicePage(self._version, response, self._solution)

    def get(self, sid):
        """
        Constructs a ServiceContext

        :param sid: The unique string that identifies the resource

        :returns: twilio.rest.chat.v1.service.ServiceContext
        :rtype: twilio.rest.chat.v1.service.ServiceContext
        """
        return ServiceContext(self._version, sid=sid, )

    def __call__(self, sid):
        """
        Constructs a ServiceContext

        :param sid: The unique string that identifies the resource

        :returns: twilio.rest.chat.v1.service.ServiceContext
        :rtype: twilio.rest.chat.v1.service.ServiceContext
        """
        return ServiceContext(self._version, sid=sid, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Chat.V1.ServiceList>'


class ServicePage(Page):

    def __init__(self, version, response, solution):
        """
        Initialize the ServicePage

        :param Version version: Version that contains the resource
        :param Response response: Response from the API

        :returns: twilio.rest.chat.v1.service.ServicePage
        :rtype: twilio.rest.chat.v1.service.ServicePage
        """
        super(ServicePage, self).__init__(version, response)

        # Path Solution
        self._solution = solution

    def get_instance(self, payload):
        """
        Build an instance of ServiceInstance

        :param dict payload: Payload response from the API

        :returns: twilio.rest.chat.v1.service.ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        return ServiceInstance(self._version, payload, )

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        return '<Twilio.Chat.V1.ServicePage>'


class ServiceContext(InstanceContext):

    def __init__(self, version, sid):
        """
        Initialize the ServiceContext

        :param Version version: Version that contains the resource
        :param sid: The unique string that identifies the resource

        :returns: twilio.rest.chat.v1.service.ServiceContext
        :rtype: twilio.rest.chat.v1.service.ServiceContext
        """
        super(ServiceContext, self).__init__(version)

        # Path Solution
        self._solution = {'sid': sid, }
        self._uri = '/Services/{sid}'.format(**self._solution)

        # Dependents
        self._channels = None
        self._roles = None
        self._users = None

    def fetch(self):
        """
        Fetch the ServiceInstance

        :returns: The fetched ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        payload = self._version.fetch(method='GET', uri=self._uri, )

        return ServiceInstance(self._version, payload, sid=self._solution['sid'], )

    def delete(self):
        """
        Deletes the ServiceInstance

        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._version.delete(method='DELETE', uri=self._uri, )

    def update(self, friendly_name=values.unset,
               default_service_role_sid=values.unset,
               default_channel_role_sid=values.unset,
               default_channel_creator_role_sid=values.unset,
               read_status_enabled=values.unset, reachability_enabled=values.unset,
               typing_indicator_timeout=values.unset,
               consumption_report_interval=values.unset,
               notifications_new_message_enabled=values.unset,
               notifications_new_message_template=values.unset,
               notifications_added_to_channel_enabled=values.unset,
               notifications_added_to_channel_template=values.unset,
               notifications_removed_from_channel_enabled=values.unset,
               notifications_removed_from_channel_template=values.unset,
               notifications_invited_to_channel_enabled=values.unset,
               notifications_invited_to_channel_template=values.unset,
               pre_webhook_url=values.unset, post_webhook_url=values.unset,
               webhook_method=values.unset, webhook_filters=values.unset,
               webhooks_on_message_send_url=values.unset,
               webhooks_on_message_send_method=values.unset,
               webhooks_on_message_update_url=values.unset,
               webhooks_on_message_update_method=values.unset,
               webhooks_on_message_remove_url=values.unset,
               webhooks_on_message_remove_method=values.unset,
               webhooks_on_channel_add_url=values.unset,
               webhooks_on_channel_add_method=values.unset,
               webhooks_on_channel_destroy_url=values.unset,
               webhooks_on_channel_destroy_method=values.unset,
               webhooks_on_channel_update_url=values.unset,
               webhooks_on_channel_update_method=values.unset,
               webhooks_on_member_add_url=values.unset,
               webhooks_on_member_add_method=values.unset,
               webhooks_on_member_remove_url=values.unset,
               webhooks_on_member_remove_method=values.unset,
               webhooks_on_message_sent_url=values.unset,
               webhooks_on_message_sent_method=values.unset,
               webhooks_on_message_updated_url=values.unset,
               webhooks_on_message_updated_method=values.unset,
               webhooks_on_message_removed_url=values.unset,
               webhooks_on_message_removed_method=values.unset,
               webhooks_on_channel_added_url=values.unset,
               webhooks_on_channel_added_method=values.unset,
               webhooks_on_channel_destroyed_url=values.unset,
               webhooks_on_channel_destroyed_method=values.unset,
               webhooks_on_channel_updated_url=values.unset,
               webhooks_on_channel_updated_method=values.unset,
               webhooks_on_member_added_url=values.unset,
               webhooks_on_member_added_method=values.unset,
               webhooks_on_member_removed_url=values.unset,
               webhooks_on_member_removed_method=values.unset,
               limits_channel_members=values.unset,
               limits_user_channels=values.unset):
        """
        Update the ServiceInstance

        :param unicode friendly_name: A string to describe the resource
        :param unicode default_service_role_sid: The service role assigned to users when they are added to the service
        :param unicode default_channel_role_sid: The channel role assigned to users when they are added to a channel
        :param unicode default_channel_creator_role_sid: The channel role assigned to a channel creator when they join a new channel
        :param bool read_status_enabled: Whether to enable the Message Consumption Horizon feature
        :param bool reachability_enabled: Whether to enable the Reachability Indicator feature for this Service instance
        :param unicode typing_indicator_timeout: How long in seconds to wait before assuming the user is no longer typing
        :param unicode consumption_report_interval: DEPRECATED
        :param bool notifications_new_message_enabled: Whether to send a notification when a new message is added to a channel
        :param unicode notifications_new_message_template: The template to use to create the notification text displayed when a new message is added to a channel
        :param bool notifications_added_to_channel_enabled: Whether to send a notification when a member is added to a channel
        :param unicode notifications_added_to_channel_template: The template to use to create the notification text displayed when a member is added to a channel
        :param bool notifications_removed_from_channel_enabled: Whether to send a notification to a user when they are removed from a channel
        :param unicode notifications_removed_from_channel_template: The template to use to create the notification text displayed to a user when they are removed
        :param bool notifications_invited_to_channel_enabled: Whether to send a notification when a user is invited to a channel
        :param unicode notifications_invited_to_channel_template: The template to use to create the notification text displayed when a user is invited to a channel
        :param unicode pre_webhook_url: The webhook URL for pre-event webhooks
        :param unicode post_webhook_url: The URL for post-event webhooks
        :param unicode webhook_method: The HTTP method  to use for both PRE and POST webhooks
        :param unicode webhook_filters: The list of WebHook events that are enabled for this Service instance
        :param unicode webhooks_on_message_send_url: The URL of the webhook to call in response to the on_message_send event
        :param unicode webhooks_on_message_send_method: The HTTP method to use when calling the webhooks.on_message_send.url
        :param unicode webhooks_on_message_update_url: The URL of the webhook to call in response to the on_message_update event
        :param unicode webhooks_on_message_update_method: The HTTP method to use when calling the webhooks.on_message_update.url
        :param unicode webhooks_on_message_remove_url: The URL of the webhook to call in response to the on_message_remove event
        :param unicode webhooks_on_message_remove_method: The HTTP method to use when calling the webhooks.on_message_remove.url
        :param unicode webhooks_on_channel_add_url: The URL of the webhook to call in response to the on_channel_add event
        :param unicode webhooks_on_channel_add_method: The HTTP method to use when calling the webhooks.on_channel_add.url
        :param unicode webhooks_on_channel_destroy_url: The URL of the webhook to call in response to the on_channel_destroy event
        :param unicode webhooks_on_channel_destroy_method: The HTTP method to use when calling the webhooks.on_channel_destroy.url
        :param unicode webhooks_on_channel_update_url: The URL of the webhook to call in response to the on_channel_update event
        :param unicode webhooks_on_channel_update_method: The HTTP method to use when calling the webhooks.on_channel_update.url
        :param unicode webhooks_on_member_add_url: The URL of the webhook to call in response to the on_member_add event
        :param unicode webhooks_on_member_add_method: The HTTP method to use when calling the webhooks.on_member_add.url
        :param unicode webhooks_on_member_remove_url: The URL of the webhook to call in response to the on_member_remove event
        :param unicode webhooks_on_member_remove_method: The HTTP method to use when calling the webhooks.on_member_remove.url
        :param unicode webhooks_on_message_sent_url: The URL of the webhook to call in response to the on_message_sent event
        :param unicode webhooks_on_message_sent_method: The URL of the webhook to call in response to the on_message_sent event
        :param unicode webhooks_on_message_updated_url: The URL of the webhook to call in response to the on_message_updated event
        :param unicode webhooks_on_message_updated_method: The HTTP method to use when calling the webhooks.on_message_updated.url
        :param unicode webhooks_on_message_removed_url: The URL of the webhook to call in response to the on_message_removed event
        :param unicode webhooks_on_message_removed_method: The HTTP method to use when calling the webhooks.on_message_removed.url
        :param unicode webhooks_on_channel_added_url: The URL of the webhook to call in response to the on_channel_added event
        :param unicode webhooks_on_channel_added_method: The URL of the webhook to call in response to the on_channel_added event
        :param unicode webhooks_on_channel_destroyed_url: The URL of the webhook to call in response to the on_channel_added event
        :param unicode webhooks_on_channel_destroyed_method: The HTTP method to use when calling the webhooks.on_channel_destroyed.url
        :param unicode webhooks_on_channel_updated_url: he URL of the webhook to call in response to the on_channel_updated event
        :param unicode webhooks_on_channel_updated_method: The HTTP method to use when calling the webhooks.on_channel_updated.url
        :param unicode webhooks_on_member_added_url: The URL of the webhook to call in response to the on_channel_updated event
        :param unicode webhooks_on_member_added_method: he HTTP method to use when calling the webhooks.on_channel_updated.url
        :param unicode webhooks_on_member_removed_url: The URL of the webhook to call in response to the on_member_removed event
        :param unicode webhooks_on_member_removed_method: The HTTP method to use when calling the webhooks.on_member_removed.url
        :param unicode limits_channel_members: The maximum number of Members that can be added to Channels within this Service
        :param unicode limits_user_channels: The maximum number of Channels Users can be a Member of within this Service

        :returns: The updated ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        data = values.of({
            'FriendlyName': friendly_name,
            'DefaultServiceRoleSid': default_service_role_sid,
            'DefaultChannelRoleSid': default_channel_role_sid,
            'DefaultChannelCreatorRoleSid': default_channel_creator_role_sid,
            'ReadStatusEnabled': read_status_enabled,
            'ReachabilityEnabled': reachability_enabled,
            'TypingIndicatorTimeout': typing_indicator_timeout,
            'ConsumptionReportInterval': consumption_report_interval,
            'Notifications.NewMessage.Enabled': notifications_new_message_enabled,
            'Notifications.NewMessage.Template': notifications_new_message_template,
            'Notifications.AddedToChannel.Enabled': notifications_added_to_channel_enabled,
            'Notifications.AddedToChannel.Template': notifications_added_to_channel_template,
            'Notifications.RemovedFromChannel.Enabled': notifications_removed_from_channel_enabled,
            'Notifications.RemovedFromChannel.Template': notifications_removed_from_channel_template,
            'Notifications.InvitedToChannel.Enabled': notifications_invited_to_channel_enabled,
            'Notifications.InvitedToChannel.Template': notifications_invited_to_channel_template,
            'PreWebhookUrl': pre_webhook_url,
            'PostWebhookUrl': post_webhook_url,
            'WebhookMethod': webhook_method,
            'WebhookFilters': serialize.map(webhook_filters, lambda e: e),
            'Webhooks.OnMessageSend.Url': webhooks_on_message_send_url,
            'Webhooks.OnMessageSend.Method': webhooks_on_message_send_method,
            'Webhooks.OnMessageUpdate.Url': webhooks_on_message_update_url,
            'Webhooks.OnMessageUpdate.Method': webhooks_on_message_update_method,
            'Webhooks.OnMessageRemove.Url': webhooks_on_message_remove_url,
            'Webhooks.OnMessageRemove.Method': webhooks_on_message_remove_method,
            'Webhooks.OnChannelAdd.Url': webhooks_on_channel_add_url,
            'Webhooks.OnChannelAdd.Method': webhooks_on_channel_add_method,
            'Webhooks.OnChannelDestroy.Url': webhooks_on_channel_destroy_url,
            'Webhooks.OnChannelDestroy.Method': webhooks_on_channel_destroy_method,
            'Webhooks.OnChannelUpdate.Url': webhooks_on_channel_update_url,
            'Webhooks.OnChannelUpdate.Method': webhooks_on_channel_update_method,
            'Webhooks.OnMemberAdd.Url': webhooks_on_member_add_url,
            'Webhooks.OnMemberAdd.Method': webhooks_on_member_add_method,
            'Webhooks.OnMemberRemove.Url': webhooks_on_member_remove_url,
            'Webhooks.OnMemberRemove.Method': webhooks_on_member_remove_method,
            'Webhooks.OnMessageSent.Url': webhooks_on_message_sent_url,
            'Webhooks.OnMessageSent.Method': webhooks_on_message_sent_method,
            'Webhooks.OnMessageUpdated.Url': webhooks_on_message_updated_url,
            'Webhooks.OnMessageUpdated.Method': webhooks_on_message_updated_method,
            'Webhooks.OnMessageRemoved.Url': webhooks_on_message_removed_url,
            'Webhooks.OnMessageRemoved.Method': webhooks_on_message_removed_method,
            'Webhooks.OnChannelAdded.Url': webhooks_on_channel_added_url,
            'Webhooks.OnChannelAdded.Method': webhooks_on_channel_added_method,
            'Webhooks.OnChannelDestroyed.Url': webhooks_on_channel_destroyed_url,
            'Webhooks.OnChannelDestroyed.Method': webhooks_on_channel_destroyed_method,
            'Webhooks.OnChannelUpdated.Url': webhooks_on_channel_updated_url,
            'Webhooks.OnChannelUpdated.Method': webhooks_on_channel_updated_method,
            'Webhooks.OnMemberAdded.Url': webhooks_on_member_added_url,
            'Webhooks.OnMemberAdded.Method': webhooks_on_member_added_method,
            'Webhooks.OnMemberRemoved.Url': webhooks_on_member_removed_url,
            'Webhooks.OnMemberRemoved.Method': webhooks_on_member_removed_method,
            'Limits.ChannelMembers': limits_channel_members,
            'Limits.UserChannels': limits_user_channels,
        })

        payload = self._version.update(method='POST', uri=self._uri, data=data, )

        return ServiceInstance(self._version, payload, sid=self._solution['sid'], )

    @property
    def channels(self):
        """
        Access the channels

        :returns: twilio.rest.chat.v1.service.channel.ChannelList
        :rtype: twilio.rest.chat.v1.service.channel.ChannelList
        """
        if self._channels is None:
            self._channels = ChannelList(self._version, service_sid=self._solution['sid'], )
        return self._channels

    @property
    def roles(self):
        """
        Access the roles

        :returns: twilio.rest.chat.v1.service.role.RoleList
        :rtype: twilio.rest.chat.v1.service.role.RoleList
        """
        if self._roles is None:
            self._roles = RoleList(self._version, service_sid=self._solution['sid'], )
        return self._roles

    @property
    def users(self):
        """
        Access the users

        :returns: twilio.rest.chat.v1.service.user.UserList
        :rtype: twilio.rest.chat.v1.service.user.UserList
        """
        if self._users is None:
            self._users = UserList(self._version, service_sid=self._solution['sid'], )
        return self._users

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Chat.V1.ServiceContext {}>'.format(context)


class ServiceInstance(InstanceResource):

    def __init__(self, version, payload, sid=None):
        """
        Initialize the ServiceInstance

        :returns: twilio.rest.chat.v1.service.ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        super(ServiceInstance, self).__init__(version)

        # Marshaled Properties
        self._properties = {
            'sid': payload.get('sid'),
            'account_sid': payload.get('account_sid'),
            'friendly_name': payload.get('friendly_name'),
            'date_created': deserialize.iso8601_datetime(payload.get('date_created')),
            'date_updated': deserialize.iso8601_datetime(payload.get('date_updated')),
            'default_service_role_sid': payload.get('default_service_role_sid'),
            'default_channel_role_sid': payload.get('default_channel_role_sid'),
            'default_channel_creator_role_sid': payload.get('default_channel_creator_role_sid'),
            'read_status_enabled': payload.get('read_status_enabled'),
            'reachability_enabled': payload.get('reachability_enabled'),
            'typing_indicator_timeout': deserialize.integer(payload.get('typing_indicator_timeout')),
            'consumption_report_interval': deserialize.integer(payload.get('consumption_report_interval')),
            'limits': payload.get('limits'),
            'webhooks': payload.get('webhooks'),
            'pre_webhook_url': payload.get('pre_webhook_url'),
            'post_webhook_url': payload.get('post_webhook_url'),
            'webhook_method': payload.get('webhook_method'),
            'webhook_filters': payload.get('webhook_filters'),
            'notifications': payload.get('notifications'),
            'url': payload.get('url'),
            'links': payload.get('links'),
        }

        # Context
        self._context = None
        self._solution = {'sid': sid or self._properties['sid'], }

    @property
    def _proxy(self):
        """
        Generate an instance context for the instance, the context is capable of
        performing various actions.  All instance actions are proxied to the context

        :returns: ServiceContext for this ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceContext
        """
        if self._context is None:
            self._context = ServiceContext(self._version, sid=self._solution['sid'], )
        return self._context

    @property
    def sid(self):
        """
        :returns: The unique string that identifies the resource
        :rtype: unicode
        """
        return self._properties['sid']

    @property
    def account_sid(self):
        """
        :returns: The SID of the Account that created the resource
        :rtype: unicode
        """
        return self._properties['account_sid']

    @property
    def friendly_name(self):
        """
        :returns: The string that you assigned to describe the resource
        :rtype: unicode
        """
        return self._properties['friendly_name']

    @property
    def date_created(self):
        """
        :returns: The RFC 2822 date and time in GMT when the resource was created
        :rtype: datetime
        """
        return self._properties['date_created']

    @property
    def date_updated(self):
        """
        :returns: The RFC 2822 date and time in GMT when the resource was last updated
        :rtype: datetime
        """
        return self._properties['date_updated']

    @property
    def default_service_role_sid(self):
        """
        :returns: The service role assigned to users when they are added to the service
        :rtype: unicode
        """
        return self._properties['default_service_role_sid']

    @property
    def default_channel_role_sid(self):
        """
        :returns: The channel role assigned to users when they are added to a channel
        :rtype: unicode
        """
        return self._properties['default_channel_role_sid']

    @property
    def default_channel_creator_role_sid(self):
        """
        :returns: The channel role assigned to a channel creator when they join a new channel
        :rtype: unicode
        """
        return self._properties['default_channel_creator_role_sid']

    @property
    def read_status_enabled(self):
        """
        :returns: Whether the Message Consumption Horizon feature is enabled
        :rtype: bool
        """
        return self._properties['read_status_enabled']

    @property
    def reachability_enabled(self):
        """
        :returns: Whether the Reachability Indicator feature is enabled for this Service instance
        :rtype: bool
        """
        return self._properties['reachability_enabled']

    @property
    def typing_indicator_timeout(self):
        """
        :returns: How long in seconds to wait before assuming the user is no longer typing
        :rtype: unicode
        """
        return self._properties['typing_indicator_timeout']

    @property
    def consumption_report_interval(self):
        """
        :returns: DEPRECATED
        :rtype: unicode
        """
        return self._properties['consumption_report_interval']

    @property
    def limits(self):
        """
        :returns: An object that describes the limits of the service instance
        :rtype: dict
        """
        return self._properties['limits']

    @property
    def webhooks(self):
        """
        :returns: An object that contains information about the webhooks configured for this service
        :rtype: dict
        """
        return self._properties['webhooks']

    @property
    def pre_webhook_url(self):
        """
        :returns: The webhook URL for pre-event webhooks
        :rtype: unicode
        """
        return self._properties['pre_webhook_url']

    @property
    def post_webhook_url(self):
        """
        :returns: The URL for post-event webhooks
        :rtype: unicode
        """
        return self._properties['post_webhook_url']

    @property
    def webhook_method(self):
        """
        :returns: The HTTP method  to use for both PRE and POST webhooks
        :rtype: unicode
        """
        return self._properties['webhook_method']

    @property
    def webhook_filters(self):
        """
        :returns: The list of WebHook events that are enabled for this Service instance
        :rtype: unicode
        """
        return self._properties['webhook_filters']

    @property
    def notifications(self):
        """
        :returns: The notification configuration for the Service instance
        :rtype: dict
        """
        return self._properties['notifications']

    @property
    def url(self):
        """
        :returns: The absolute URL of the Service resource
        :rtype: unicode
        """
        return self._properties['url']

    @property
    def links(self):
        """
        :returns: The absolute URLs of the Service's Channels, Roles, and Users
        :rtype: unicode
        """
        return self._properties['links']

    def fetch(self):
        """
        Fetch the ServiceInstance

        :returns: The fetched ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        return self._proxy.fetch()

    def delete(self):
        """
        Deletes the ServiceInstance

        :returns: True if delete succeeds, False otherwise
        :rtype: bool
        """
        return self._proxy.delete()

    def update(self, friendly_name=values.unset,
               default_service_role_sid=values.unset,
               default_channel_role_sid=values.unset,
               default_channel_creator_role_sid=values.unset,
               read_status_enabled=values.unset, reachability_enabled=values.unset,
               typing_indicator_timeout=values.unset,
               consumption_report_interval=values.unset,
               notifications_new_message_enabled=values.unset,
               notifications_new_message_template=values.unset,
               notifications_added_to_channel_enabled=values.unset,
               notifications_added_to_channel_template=values.unset,
               notifications_removed_from_channel_enabled=values.unset,
               notifications_removed_from_channel_template=values.unset,
               notifications_invited_to_channel_enabled=values.unset,
               notifications_invited_to_channel_template=values.unset,
               pre_webhook_url=values.unset, post_webhook_url=values.unset,
               webhook_method=values.unset, webhook_filters=values.unset,
               webhooks_on_message_send_url=values.unset,
               webhooks_on_message_send_method=values.unset,
               webhooks_on_message_update_url=values.unset,
               webhooks_on_message_update_method=values.unset,
               webhooks_on_message_remove_url=values.unset,
               webhooks_on_message_remove_method=values.unset,
               webhooks_on_channel_add_url=values.unset,
               webhooks_on_channel_add_method=values.unset,
               webhooks_on_channel_destroy_url=values.unset,
               webhooks_on_channel_destroy_method=values.unset,
               webhooks_on_channel_update_url=values.unset,
               webhooks_on_channel_update_method=values.unset,
               webhooks_on_member_add_url=values.unset,
               webhooks_on_member_add_method=values.unset,
               webhooks_on_member_remove_url=values.unset,
               webhooks_on_member_remove_method=values.unset,
               webhooks_on_message_sent_url=values.unset,
               webhooks_on_message_sent_method=values.unset,
               webhooks_on_message_updated_url=values.unset,
               webhooks_on_message_updated_method=values.unset,
               webhooks_on_message_removed_url=values.unset,
               webhooks_on_message_removed_method=values.unset,
               webhooks_on_channel_added_url=values.unset,
               webhooks_on_channel_added_method=values.unset,
               webhooks_on_channel_destroyed_url=values.unset,
               webhooks_on_channel_destroyed_method=values.unset,
               webhooks_on_channel_updated_url=values.unset,
               webhooks_on_channel_updated_method=values.unset,
               webhooks_on_member_added_url=values.unset,
               webhooks_on_member_added_method=values.unset,
               webhooks_on_member_removed_url=values.unset,
               webhooks_on_member_removed_method=values.unset,
               limits_channel_members=values.unset,
               limits_user_channels=values.unset):
        """
        Update the ServiceInstance

        :param unicode friendly_name: A string to describe the resource
        :param unicode default_service_role_sid: The service role assigned to users when they are added to the service
        :param unicode default_channel_role_sid: The channel role assigned to users when they are added to a channel
        :param unicode default_channel_creator_role_sid: The channel role assigned to a channel creator when they join a new channel
        :param bool read_status_enabled: Whether to enable the Message Consumption Horizon feature
        :param bool reachability_enabled: Whether to enable the Reachability Indicator feature for this Service instance
        :param unicode typing_indicator_timeout: How long in seconds to wait before assuming the user is no longer typing
        :param unicode consumption_report_interval: DEPRECATED
        :param bool notifications_new_message_enabled: Whether to send a notification when a new message is added to a channel
        :param unicode notifications_new_message_template: The template to use to create the notification text displayed when a new message is added to a channel
        :param bool notifications_added_to_channel_enabled: Whether to send a notification when a member is added to a channel
        :param unicode notifications_added_to_channel_template: The template to use to create the notification text displayed when a member is added to a channel
        :param bool notifications_removed_from_channel_enabled: Whether to send a notification to a user when they are removed from a channel
        :param unicode notifications_removed_from_channel_template: The template to use to create the notification text displayed to a user when they are removed
        :param bool notifications_invited_to_channel_enabled: Whether to send a notification when a user is invited to a channel
        :param unicode notifications_invited_to_channel_template: The template to use to create the notification text displayed when a user is invited to a channel
        :param unicode pre_webhook_url: The webhook URL for pre-event webhooks
        :param unicode post_webhook_url: The URL for post-event webhooks
        :param unicode webhook_method: The HTTP method  to use for both PRE and POST webhooks
        :param unicode webhook_filters: The list of WebHook events that are enabled for this Service instance
        :param unicode webhooks_on_message_send_url: The URL of the webhook to call in response to the on_message_send event
        :param unicode webhooks_on_message_send_method: The HTTP method to use when calling the webhooks.on_message_send.url
        :param unicode webhooks_on_message_update_url: The URL of the webhook to call in response to the on_message_update event
        :param unicode webhooks_on_message_update_method: The HTTP method to use when calling the webhooks.on_message_update.url
        :param unicode webhooks_on_message_remove_url: The URL of the webhook to call in response to the on_message_remove event
        :param unicode webhooks_on_message_remove_method: The HTTP method to use when calling the webhooks.on_message_remove.url
        :param unicode webhooks_on_channel_add_url: The URL of the webhook to call in response to the on_channel_add event
        :param unicode webhooks_on_channel_add_method: The HTTP method to use when calling the webhooks.on_channel_add.url
        :param unicode webhooks_on_channel_destroy_url: The URL of the webhook to call in response to the on_channel_destroy event
        :param unicode webhooks_on_channel_destroy_method: The HTTP method to use when calling the webhooks.on_channel_destroy.url
        :param unicode webhooks_on_channel_update_url: The URL of the webhook to call in response to the on_channel_update event
        :param unicode webhooks_on_channel_update_method: The HTTP method to use when calling the webhooks.on_channel_update.url
        :param unicode webhooks_on_member_add_url: The URL of the webhook to call in response to the on_member_add event
        :param unicode webhooks_on_member_add_method: The HTTP method to use when calling the webhooks.on_member_add.url
        :param unicode webhooks_on_member_remove_url: The URL of the webhook to call in response to the on_member_remove event
        :param unicode webhooks_on_member_remove_method: The HTTP method to use when calling the webhooks.on_member_remove.url
        :param unicode webhooks_on_message_sent_url: The URL of the webhook to call in response to the on_message_sent event
        :param unicode webhooks_on_message_sent_method: The URL of the webhook to call in response to the on_message_sent event
        :param unicode webhooks_on_message_updated_url: The URL of the webhook to call in response to the on_message_updated event
        :param unicode webhooks_on_message_updated_method: The HTTP method to use when calling the webhooks.on_message_updated.url
        :param unicode webhooks_on_message_removed_url: The URL of the webhook to call in response to the on_message_removed event
        :param unicode webhooks_on_message_removed_method: The HTTP method to use when calling the webhooks.on_message_removed.url
        :param unicode webhooks_on_channel_added_url: The URL of the webhook to call in response to the on_channel_added event
        :param unicode webhooks_on_channel_added_method: The URL of the webhook to call in response to the on_channel_added event
        :param unicode webhooks_on_channel_destroyed_url: The URL of the webhook to call in response to the on_channel_added event
        :param unicode webhooks_on_channel_destroyed_method: The HTTP method to use when calling the webhooks.on_channel_destroyed.url
        :param unicode webhooks_on_channel_updated_url: he URL of the webhook to call in response to the on_channel_updated event
        :param unicode webhooks_on_channel_updated_method: The HTTP method to use when calling the webhooks.on_channel_updated.url
        :param unicode webhooks_on_member_added_url: The URL of the webhook to call in response to the on_channel_updated event
        :param unicode webhooks_on_member_added_method: he HTTP method to use when calling the webhooks.on_channel_updated.url
        :param unicode webhooks_on_member_removed_url: The URL of the webhook to call in response to the on_member_removed event
        :param unicode webhooks_on_member_removed_method: The HTTP method to use when calling the webhooks.on_member_removed.url
        :param unicode limits_channel_members: The maximum number of Members that can be added to Channels within this Service
        :param unicode limits_user_channels: The maximum number of Channels Users can be a Member of within this Service

        :returns: The updated ServiceInstance
        :rtype: twilio.rest.chat.v1.service.ServiceInstance
        """
        return self._proxy.update(
            friendly_name=friendly_name,
            default_service_role_sid=default_service_role_sid,
            default_channel_role_sid=default_channel_role_sid,
            default_channel_creator_role_sid=default_channel_creator_role_sid,
            read_status_enabled=read_status_enabled,
            reachability_enabled=reachability_enabled,
            typing_indicator_timeout=typing_indicator_timeout,
            consumption_report_interval=consumption_report_interval,
            notifications_new_message_enabled=notifications_new_message_enabled,
            notifications_new_message_template=notifications_new_message_template,
            notifications_added_to_channel_enabled=notifications_added_to_channel_enabled,
            notifications_added_to_channel_template=notifications_added_to_channel_template,
            notifications_removed_from_channel_enabled=notifications_removed_from_channel_enabled,
            notifications_removed_from_channel_template=notifications_removed_from_channel_template,
            notifications_invited_to_channel_enabled=notifications_invited_to_channel_enabled,
            notifications_invited_to_channel_template=notifications_invited_to_channel_template,
            pre_webhook_url=pre_webhook_url,
            post_webhook_url=post_webhook_url,
            webhook_method=webhook_method,
            webhook_filters=webhook_filters,
            webhooks_on_message_send_url=webhooks_on_message_send_url,
            webhooks_on_message_send_method=webhooks_on_message_send_method,
            webhooks_on_message_update_url=webhooks_on_message_update_url,
            webhooks_on_message_update_method=webhooks_on_message_update_method,
            webhooks_on_message_remove_url=webhooks_on_message_remove_url,
            webhooks_on_message_remove_method=webhooks_on_message_remove_method,
            webhooks_on_channel_add_url=webhooks_on_channel_add_url,
            webhooks_on_channel_add_method=webhooks_on_channel_add_method,
            webhooks_on_channel_destroy_url=webhooks_on_channel_destroy_url,
            webhooks_on_channel_destroy_method=webhooks_on_channel_destroy_method,
            webhooks_on_channel_update_url=webhooks_on_channel_update_url,
            webhooks_on_channel_update_method=webhooks_on_channel_update_method,
            webhooks_on_member_add_url=webhooks_on_member_add_url,
            webhooks_on_member_add_method=webhooks_on_member_add_method,
            webhooks_on_member_remove_url=webhooks_on_member_remove_url,
            webhooks_on_member_remove_method=webhooks_on_member_remove_method,
            webhooks_on_message_sent_url=webhooks_on_message_sent_url,
            webhooks_on_message_sent_method=webhooks_on_message_sent_method,
            webhooks_on_message_updated_url=webhooks_on_message_updated_url,
            webhooks_on_message_updated_method=webhooks_on_message_updated_method,
            webhooks_on_message_removed_url=webhooks_on_message_removed_url,
            webhooks_on_message_removed_method=webhooks_on_message_removed_method,
            webhooks_on_channel_added_url=webhooks_on_channel_added_url,
            webhooks_on_channel_added_method=webhooks_on_channel_added_method,
            webhooks_on_channel_destroyed_url=webhooks_on_channel_destroyed_url,
            webhooks_on_channel_destroyed_method=webhooks_on_channel_destroyed_method,
            webhooks_on_channel_updated_url=webhooks_on_channel_updated_url,
            webhooks_on_channel_updated_method=webhooks_on_channel_updated_method,
            webhooks_on_member_added_url=webhooks_on_member_added_url,
            webhooks_on_member_added_method=webhooks_on_member_added_method,
            webhooks_on_member_removed_url=webhooks_on_member_removed_url,
            webhooks_on_member_removed_method=webhooks_on_member_removed_method,
            limits_channel_members=limits_channel_members,
            limits_user_channels=limits_user_channels,
        )

    @property
    def channels(self):
        """
        Access the channels

        :returns: twilio.rest.chat.v1.service.channel.ChannelList
        :rtype: twilio.rest.chat.v1.service.channel.ChannelList
        """
        return self._proxy.channels

    @property
    def roles(self):
        """
        Access the roles

        :returns: twilio.rest.chat.v1.service.role.RoleList
        :rtype: twilio.rest.chat.v1.service.role.RoleList
        """
        return self._proxy.roles

    @property
    def users(self):
        """
        Access the users

        :returns: twilio.rest.chat.v1.service.user.UserList
        :rtype: twilio.rest.chat.v1.service.user.UserList
        """
        return self._proxy.users

    def __repr__(self):
        """
        Provide a friendly representation

        :returns: Machine friendly representation
        :rtype: str
        """
        context = ' '.join('{}={}'.format(k, v) for k, v in self._solution.items())
        return '<Twilio.Chat.V1.ServiceInstance {}>'.format(context)
